from astropy.table import Table
from astropy.units import UnitsWarning
from ..path import tree_path
from functools import partial
import warnings


class AuxService:

    renames = {}
    ignored_columns = []
    transforms = {}
    basename = 'AUX_SERVICE'

    def __init__(self, auxdir='/fact/aux'):
        self.path = partial(
            tree_path,
            base_dir=auxdir,
            suffix='.' + self.basename + '.fits')

    @classmethod
    def read_file(cls, filename):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=UnitsWarning)
            table = Table.read(filename)

        for column in table.columns.keys():
            if column in cls.ignored_columns:
                table.remove_column(column)

            elif column in cls.renames:
                table[column].name = cls.renames[column]

        for column in table.columns.keys():
            shape = table[column].shape
            if len(shape) > 1:
                for i in range(shape[1]):
                    table[column + '_{}'.format(i)] = table[column][:, i]
                table.remove_column(column)

        df = table.to_pandas()

        for key, transform in cls.transforms.items():
            df[key] = transform(df[key])

        return df

    def read_date(self, date):
        return self.read_file(self.path(int('{:%Y%m%d}'.format(date))))



import numpy as np

from .factdb import (
    read_into_dataframe,
    AnalysisResultsRunLP as QLA,
    RunInfo,
    Source
)


def get_qla_data(
        first_night=None,
        last_night=None,
        sources=None,
        database_engine=None
        ):
    '''
    Request QLA results from our database

    first_night: int or None
        If given, first night to query as FACT night integer.
    last_night: int or None
        If given, last night to query as FACT night integer.
    sources: iterable[str]
        If given, only these sources will be requested.
        Names have to match Source.fSourceName in our db.
    database_engine: sqlalchmey.Engine
        If given, the connection to use for the query.
        Else, `fact.credentials.create_factdb_engine` will be used to create it.
    '''

    query = QLA.select(
        QLA.frunid.alias('run_id'),
        QLA.fnight.alias('night'),
        QLA.fnumexcevts.alias('n_excess'),
        QLA.fnumsigevts.alias('n_on'),
        (QLA.fnumbgevts * 5).alias('n_off'),
        QLA.fontimeaftercuts.alias('ontime'),
        RunInfo.frunstart.alias('run_start'),
        RunInfo.frunstop.alias('run_stop'),
        Source.fsourcename.alias('source'),
        Source.fsourcekey.alias('source_key'),
    )

    on = (RunInfo.fnight == QLA.fnight) & (RunInfo.frunid == QLA.frunid)
    query = query.join(RunInfo, on=on)
    query = query.join(Source, on=RunInfo.fsourcekey == Source.fsourcekey)

    if first_night is not None:
        query = query.where(QLA.fnight >= first_night)
    if last_night is not None:
        query = query.where(QLA.fnight <= last_night)

    if sources is not None:
        query = query.where(Source.fsourcename.in_(sources))

    runs = read_into_dataframe(query, engine=database_engine)

    # drop rows with NaNs from the table, these are unfinished qla results
    runs.dropna(inplace=True)

    runs.sort_values('run_start', inplace=True)

    return runs



""" library of (hopefully) useful functions for the slow data DB interface
"""
import numpy as np
import pymongo


def cursor_to_rec_array(cursor):
    """ convert a pymongo.cursor.Cursor to an numpz recarray
    """
    array = cursor_to_structured_array(cursor)
    return array.view(np.recarray)


def cursor_to_structured_array(cursor):
    """ convert a pymongo.cursor.Cursor to an numpy structured array
    """
    number_of_docs = cursor.count()
    # if number_of_docs > 1000:
    #    logging.warning("loading {} documents form database"
    #                 " .. might take a while".format(number_of_docs))

    structured_array_dtype = make_numpy_dtype_from_cursor(cursor)
    array = np.zeros(number_of_docs, structured_array_dtype)

    for counter, document in enumerate(cursor):
        for field_name in structured_array_dtype.names:
            try:
                array[field_name][counter] = document[field_name]
            except KeyError:
                array[field_name][counter] = np.nan

    return array


def make_numpy_dtype_from_cursor(cursor):
    """ infer datatype of structured array from document(s) from a cursor
    """
    collection_of_this_cursor = cursor.collection
    # get the newest entry from this collection
    # this one defines the dtype of the numpy array,
    # and we want to stick to the newest format.
    example_document = collection_of_this_cursor.find_one(
        {},
        sort=[("Time", pymongo.DESCENDING)]
    )
    example_document = cursor[0]

    if example_document is None:
        # collection is empty
        raise LookupError('associated collection of cursor is empty.')

    return make_numpy_dtype_from_document(example_document)


def make_numpy_dtype_from_document(doc):
    """ infer datatype of structured array from document
    """
    list_of_names_n_types = []
    for field_name in doc:
        if '_id' in field_name:
            continue
        element = doc[field_name]
        element_array = np.array(element)
        list_of_names_n_types.append(
            (str(field_name),
             element_array.dtype.str,
             element_array.shape)
        )

    return np.dtype(list_of_names_n_types)



import pandas as pd
import peewee
from .models import RunInfo, Source, RunType
from ..credentials import create_factdb_engine
from ..time import night_integer


SECOND = peewee.SQL('SECOND')
run_duration = peewee.fn.TIMESTAMPDIFF(
    SECOND, RunInfo.frunstart, RunInfo.frunstop
)
ontime = (run_duration * RunInfo.feffectiveon)


def read_into_dataframe(query, engine=None):
    ''' read the result of a peewee query object into a pandas DataFrame '''
    engine = engine or create_factdb_engine()
    sql, params = query.sql()
    
    with engine.connect() as conn:
        df = pd.read_sql_query(sql, conn, params=params)

    return df


def get_correct_ontime(start=None, end=None, engine=None):
    '''
    The database field fOnTime underestimates the real ontime by about 5 seconds
    because of how the number is calculated from the FTM auxfiles.
    A better estimate can be obtained by taking (fRunStop - fRunStart) * fEffectiveOn.

    Parameters
    ----------
    start : int or datetime.date
        First night to select, either in fact int format or as date
    end : int or datetime.date
        Last night to select, either in fact int format or as date
    engine: sqlalchemy.Engine
        The engine connected to the database.
        If None, fact.credentials.create_factdb_engine will be used to create one.

    Source: D. Neise, A. Biland. Also see github.com/dneise/about_fact_ontime
    '''

    query = RunInfo.select(
        RunInfo.fnight.alias('night'),
        RunInfo.frunid.alias('run_id'),
        RunInfo.frunstart.alias('start'),
        RunInfo.frunstart.alias('stop'),
        ontime.alias('ontime'),
    )

    if start is not None:
        start = night_integer(start) if not isinstance(start, int) else start
        query = query.where(RunInfo.fnight >= start)

    if end is not None:
        end = night_integer(end) if not isinstance(end, int) else end
        query = query.where(RunInfo.fnight <= end)

    df = read_into_dataframe(query, engine=engine)

    return df


def get_ontime_by_source_and_runtype(engine=None):
    query = (
        RunInfo
        .select(
            peewee.fn.SUM(ontime).alias('ontime'),
            Source.fsourcename.alias('source'),
            RunType.fruntypename.alias('runtype')
        )
        .join(Source, on=Source.fsourcekey == RunInfo.fsourcekey)
        .switch(RunInfo)
        .join(RunType, on=RunType.fruntypekey == RunInfo.fruntypekey)
        .group_by(Source.fsourcename, RunType.fruntypename)
    )
    df = read_into_dataframe(query, engine or create_factdb_engine())
    df.set_index(['source', 'runtype'], inplace=True)

    return df


def get_ontime_by_source(runtype=None, engine=None):
    query = (
        RunInfo
        .select(
            peewee.fn.SUM(ontime).alias('ontime'),
            Source.fsourcename.alias('source'),
        )
        .join(Source, on=Source.fsourcekey == RunInfo.fsourcekey)
        .switch(RunInfo)
        .join(RunType, on=RunType.fruntypekey == RunInfo.fruntypekey)
    )
    if runtype is not None:
        query = query.where(RunType.fruntypename == runtype)

    query = query.group_by(Source.fsourcename)

    df = read_into_dataframe(query, engine or create_factdb_engine())
    df.set_index('source', inplace=True)

    return df



'''
The varoius trigger types of the FACT camera.

For a full overview of the FACT trigger types, see the 
[Phd of Patrick Vogler, table 4.3.b, page 93]
(http://e-collection.library.ethz.ch/eserv/eth:48381/eth-48381-02.pdf)

The trigger type IDs also encode the N-out-of-4 trigger patten logic, but here 
are only trigger type IDs for N=1 as it is the most common.
'''

#: Self triggered, these events are likely to contain photon clusters.
PHYSICS = 4
#: These events are likely to contain only night sky background photons.
PEDESTAL = 1024
#: The external light pulser which is located in the center of the reflector 
#: dish.
LIGHT_PULSER_EXTERNAL = 260
#: The internal lightpulser was deactivated in May 2014.
LIGHT_PULSER_INTERNAL = 512
#: Not sure, seems to be for the DRS4 time calibration.
TIME_CALIBRATION = 33792
#: These events are likely to contain only night sky background photons.
#: Here the GPS module is used as reference clock.  The GPS is sometimes 
#: connected to EXT1 and sometimes to EXT2.
EXT1 = 1
#: See EXT1.
EXT2 = 2


''' FACT slow database frontent

provided for your convenience.
'''
import numpy as np
from pymongo import MongoClient
from . import settings
from . import tools
from . import correlation


__all__ = ['connect', 'correlation']


class ServiceField(object):
    ''' represents one Field of a Service
    '''

    def __init__(self, collection, key):
        self.__collection = collection
        self.__key = key

    def from_until(self, start, end):
        ''' retrieve field from start to end date as numpy array.
        '''
        cursor = self.__collection.find({'Time': {'$gte': start, '$lt': end}})
        return tools.cursor_to_structured_array(cursor)


class Service(object):
    ''' represents a FACT slow data service
    '''

    def __init__(self, collection):
        self.__collection = collection
        self.__get_service_field_names(collection)
        for key in self.__keys:
            setattr(
                self,
                key,
                ServiceField(collection, key))

        self.__keys = None

    def __get_service_field_names(self, coll):
        ''' retrieve the field names of this service

        by  using one example decument.
        '''
        self.__keys = coll.find_one()
        del self.__keys['_id']

    def from_until(self, start, end, sample_period=None, skip=None, fields=None):
        ''' retrieve service from start to end date as numpy array.

        start : starttime in FJD
        end   : end time in FJD

        keyword arguments:

        sample_period : time(in days) into which the request should be splitted --> skip
        skip   : number of periods, which should be skipped during request
        fields : list of field names to be requested [default: all fields]
        '''
        if sample_period is not None:
            sample_boundaries = np.arange(start, end, sample_period)
        else:
            sample_boundaries = np.array([start, end])

        samples_start = sample_boundaries[0:-1:skip]
        samples_end = sample_boundaries[1::skip]
        dict_list = []
        for _start, _end in zip(samples_start, samples_end):
            start_stop_dict = {'Time': {
                '$gte': float(_start),
                '$lt': float(_end)}
            }
            dict_list.append(start_stop_dict)

        if fields is not None:
            cursor = self.__collection.find({'$or': dict_list}, fields=fields)
        else:
            cursor = self.__collection.find({'$or': dict_list})

        return tools.cursor_to_structured_array(cursor)


class AuxDataBaseFrontEnd(object):

    def __init__(self, database):
        self.database = database
        self.__service_names = None  # to be initialised in __init_service_names
        self.__fill_in_services()

    def __fill_in_services(self):
        self.__init_service_names()
        for service_name in self.__service_names:
            service = Service(self.database[service_name])
            setattr(self, service_name, service)

    def __init_service_names(self):
        self.__service_names = list()
        for collection_name in self.database.collection_names():
            if 'system.indexes' not in collection_name:
                self.__service_names.append(collection_name)


def connect(host=None, port=None, db_name=None):
    if host is None:
        host = settings.host

    if port is None:
        port = settings.port

    if db_name is None:
        db_name = settings.db_name

    client = MongoClient(host, port)
    return AuxDataBaseFrontEnd(getattr(client, db_name))



'''
This module contains functions to plot fact data into the camera view.

The Viewer class starts a GUI with tkinter, that let's you click through
events. You will only have access to the Viewer if you have installed
matplotlib with tcl/tk support

Currently these functions only work with shape (num_events, 1440), so
on a pixel bases
'''
import warnings
from .core import mark_pixel, camera, pixelids
from .analysis import plot_excess_rate

__all__ = [
    'camera',
    'mark_pixel',
    'pixelids',
    'plot_excess_rate',
]

try:
    from .viewer import Viewer
    __all__.append('Viewer')
except ImportError:
    warnings.warn('Matplotlib was build without tkagg support,\n'
                  'the Viewer will not be available')



from .time import fjd, iso2dt, run2dt, facttime, night, night_integer, datestr
from . import plotting
from . import auxservices
from pkg_resources import resource_string

__version__ = resource_string('fact', 'VERSION').decode().strip()


__all__ = [
    'fjd',
    'iso2dt',
    'run2dt',
    'facttime',
    'night',
    'night_integer',
    'datestr',
    'plotting',
    'auxservices',
]



import numpy as np
import pandas as pd
import re

from .statistics import li_ma_significance

default_theta_off_keys = tuple('theta_deg_off_{}'.format(i) for i in range(1, 6))
default_prediction_off_keys = tuple(
    'gamma_prediction_off_{}'.format(i) for i in range(1, 6)
)


off_key_re = re.compile('([a-zA-z1-9]+_deg)_off_([0-9])?')


def calc_run_summary_source_independent(
        events, runs,
        prediction_threshold,
        theta2_cut,
        prediction_key='gamma_prediction',
        theta_key='theta_deg',
        theta_off_keys=default_theta_off_keys,
        ):

    '''
    Calculate run summaries for the given theta^2 and signal prediction cuts.
    This function requires, that no source dependent features,
    like Theta were used in the classification.

    Parameters
    ----------
    events: pd.DataFrame
        DataFrame with event data, needs to contain the columns
        `'night'`, `'run'`, `theta_key` and the `theta_off_keys`
    prediction_threshold: float
        Threshold for the classifier prediction
    theta2_cut: float
        Selection cut for theta^2 in deg^2
    prediction_key: str:
        Key to the classifier prediction
    theta_key: str
        Column name of the column containing theta in degree
    theta_off_keys: list[str]
        Column names of the column containing theta  in degree
        for all off regions
    '''

    runs = runs.set_index(['night', 'run_id'])
    runs.sort_index(inplace=True)

    # apply prediction threshold cut
    selected = events.query(
        '{} >= {}'.format(prediction_key, prediction_threshold)
    )

    on_data, off_data = split_on_off_source_independent(
        selected, theta2_cut, theta_key, theta_off_keys
    )

    alpha = 1 / len(theta_off_keys)

    runs['n_on'] = on_data.groupby(['night', 'run_id']).size()
    runs['n_on'].fillna(0, inplace=True)

    runs['n_off'] = off_data.groupby(['night', 'run_id']).size()
    runs['n_off'].fillna(0, inplace=True)

    runs['n_excess'] = runs['n_on'] - alpha * runs['n_off']
    runs['n_excess_err'] = np.sqrt(runs['n_on'] + alpha**2 * runs['n_off'])

    runs['excess_rate_per_h'] = runs['n_excess'] / runs['ontime'] / 3600
    runs['excess_rate_per_h_err'] = runs['n_excess_err'] / runs['ontime'] / 3600

    runs['significance'] = li_ma_significance(
        runs['n_on'], runs['n_off'], alpha
    )

    runs.reset_index(inplace=True)

    return runs


def split_on_off_source_independent(
        events,
        theta2_cut,
        theta_key='theta_deg',
        theta_off_keys=default_theta_off_keys,
        ):
    '''
    Split events dataframe into on and off region

    Parameters
    ----------
    events: pd.DataFrame
        DataFrame containing event information, required are
        `theta_key` and `theta_off_keys`.
    theta2_cut: float
        Selection cut for theta^2 in deg^2
    theta_key: str
        Column name of the column containing theta in degree
    theta_off_keys: list[str]
        Column names of the column containing theta  in degree
        for all off regions
    '''
    # apply theta2_cut
    theta_cut = np.sqrt(theta2_cut)

    on_data = events.query('{} <= {}'.format(theta_key, theta_cut))

    off_dfs = []
    for region, theta_off_key in enumerate(theta_off_keys, start=1):
        off_df = events.query('{} <= {}'.format(
            theta_off_key, theta_cut)
        ).copy()

        off_df['off_region'] = region
        drop_off_columns(off_df, region, inplace=True)

        off_dfs.append(off_df)

    off_data = pd.concat(off_dfs)

    return on_data, off_data


def calc_run_summary_source_dependent(
        events, runs,
        prediction_threshold,
        on_prediction_key='gamma_prediction',
        off_prediction_keys=default_prediction_off_keys,
        ):
    '''
    Calculate run summaries for the given signal prediction cuts.
    This function needs to be used, if source dependent features like
    Theta were used for the classification.

    Parameters
    ----------
    events: pd.DataFrame
        DataFrame with event data, needs to contain the columns
        `'night'`, `'run'`, `theta_key` and the `theta_off_keys`
    prediction_threshold: float
        Threshold for the signalness prediction
    on_prediction_key: str
        Key to the classifier prediction for the on region
    off_prediction_keys: list[str]
        Iterable of keys to the classifier predictions for the off regions
    '''

    runs = runs.set_index(['night', 'run_id'])
    runs.sort_index(inplace=True)

    on_data, off_data = split_on_off_source_dependent(
        events, prediction_threshold, on_prediction_key, off_prediction_keys
    )

    alpha = 1 / len(off_prediction_keys)

    runs['n_on'] = on_data.groupby(['night', 'run_id']).size()
    runs['n_on'].fillna(0, inplace=True)

    runs['n_off'] = off_data.groupby(['night', 'run_id']).size()
    runs['n_off'].fillna(0, inplace=True)

    runs['significance'] = li_ma_significance(
        runs['n_on'], runs['n_off'], alpha
    )

    runs['n_excess'] = runs['n_on'] - alpha * runs['n_off']
    runs['n_excess_err'] = np.sqrt(runs['n_on'] + alpha**2 * runs['n_off'])

    runs['excess_rate_per_h'] = runs['n_excess'] / runs['ontime'] / 3600
    runs['excess_rate_per_h_err'] = runs['n_excess_err'] / runs['ontime'] / 3600

    runs.reset_index(inplace=True)

    return runs


def split_on_off_source_dependent(
        events,
        prediction_threshold,
        on_prediction_key='gamma_prediction',
        off_prediction_keys=default_prediction_off_keys,
        ):
    '''
    Split events dataframe into on and off region
    For the off regions, keys are renamed to their "on" equivalents
    and the "off" keys are dropped.

    Parameters
    ----------
    events: pd.DataFrame
        DataFrame containing event information, required are
        `theta_key` and `theta_off_keys`.
    prediction_threshold: float
        Threshold for the signalness prediction
    on_prediction_key: str
        Key to the classifier prediction for the on region
    off_prediction_keys: list[str]
        Iterable of keys to the classifier predictions for the off regions
    '''
    on_data = events.query('{} >= {}'.format(
        on_prediction_key, prediction_threshold)
    ).copy()

    off_dfs = []
    for region, off_key in enumerate(off_prediction_keys, start=1):
        off_df = events.query('{} >= {}'.format(
            off_key, prediction_threshold)
        ).copy()

        off_df['off_region'] = region

        off_df.drop(on_prediction_key, axis=1, inplace=True)
        off_df[on_prediction_key] = off_df[off_key]
        off_df.drop(off_key, axis=1, inplace=True)

        drop_off_columns(off_df, region, inplace=True)

        off_dfs.append(off_df)

    off_data = pd.concat(off_dfs)

    return on_data, off_data


def drop_off_columns(df, off_region, inplace=False):
    '''
    Replace the "On" column with the column
    of the respective off region.
    For example for `off_region = 1`, `theta` is replaced by
    theta_off_1 and all theta_off_<N> columns are dropped.
    Same for all other columns, containing the pattern `_off_<N>`
    '''
    if inplace is False:
        df = df.copy()

    for col in df.columns:
        m = off_key_re.match(col)
        if m:
            on_key, key_region = m.groups()
            if int(key_region) == off_region:
                df.drop(on_key, axis=1, inplace=True)
                df[on_key] = df[col]

            df.drop(col, axis=1, inplace=True)

    if inplace is False:
        return df



from os import path
import pandas as pd
import json
import h5py
import sys
import logging
import numpy as np
from copy import copy


__all__ = [
    'write_data',
    'to_native_byteorder',
    'read_data',
    'read_h5py',
    'read_h5py_chunked',
    'check_extension',
    'to_h5py',
]

log = logging.getLogger(__name__)


allowed_extensions = ('.hdf', '.hdf5', '.h5', '.json', '.jsonl', '.jsonlines', '.csv')
native_byteorder = native_byteorder = {'little': '<', 'big': '>'}[sys.byteorder]


def write_data(df, file_path, key='data', use_hp5y=False, **kwargs):

    name, extension = path.splitext(file_path)

    if extension in ['.hdf', '.hdf5', '.h5']:
        if use_hp5y is True:
            to_h5py(file_path, df, key=key, **kwargs)
        else:
            df.to_hdf(file_path, key=key, **kwargs)

    elif extension == '.json':
        df.to_json(file_path, **kwargs)

    elif extension in ('.jsonl', '.jsonline'):
        df.to_json(file_path, lines=True, orient='records', **kwargs)

    elif extension == '.csv':
        index = kwargs.pop('index', False)
        df.to_csv(file_path, index=index, **kwargs)

    else:
        raise IOError(
            'cannot write tabular data with format {}. Allowed formats: {}'.format(
                extension, allowed_extensions,
            )
        )


def to_native_byteorder(array):
    ''' Convert numpy array to native byteorder '''

    # '|' : not-applicable, '=': native
    if array.dtype.byteorder not in ('|', '=', native_byteorder):
        return array.byteswap().newbyteorder()

    return array


def read_h5py(file_path, key='data', columns=None, mode='r+'):
    '''
    Read a hdf5 file written with h5py into a dataframe

    Parameters
    ----------
    file_path: str
        file to read in
    key: str
        name of the hdf5 group to read in
    columns: iterable[str]
        Names of the datasets to read in. If not given read all 1d datasets
    '''
    with h5py.File(file_path, mode) as f:
        group = f.get(key)
        if group is None:
            raise IOError('File does not contain group "{}"'.format(key))

        # get all columns of which don't have more than one value per row
        if columns is None:
            columns = [col for col in group.keys() if group[col].ndim == 1]
        else:
            columns = copy(columns)

        df = pd.DataFrame()
        for col in columns:
            array = to_native_byteorder(group[col][:])

            # pandas cannot handle bytes, convert to str
            if array.dtype.kind == 'S':
                array = array.astype(str)

            if array.ndim == 1:
                df[col] = array
            elif array.ndim == 2:
                for i in range(array.shape[1]):
                    df[col + '_{}'.format(i)] = array[:, i]
            else:
                log.warning('Skipping column {}, not 1d or 2d'.format(col))

        if 'index' in df.columns:
            df.set_index('index', inplace=True)

    return df


def h5py_get_n_rows(file_path, key='data', mode='r+'):

    with h5py.File(file_path, mode) as f:
        group = f.get(key)

        if group is None:
            raise IOError('File does not contain group "{}"'.format(key))

        return group[next(iter(group.keys()))].shape[0]


def read_h5py_chunked(file_path, key='data', columns=None, chunksize=None, mode='r+'):
    '''
    Generator function to read from h5py hdf5 in chunks,
    returns an iterator over pandas dataframes.

    When chunksize is None, use 1 chunk
    '''
    with h5py.File(file_path, mode) as f:
        group = f.get(key)
        if group is None:
            raise IOError('File does not contain group "{}"'.format(key))

        # get all columns of which don't have more than one value per row
        if columns is None:
            columns = [col for col in group.keys() if group[col].ndim == 1]
        else:
            columns = copy(columns)

        n_rows = h5py_get_n_rows(file_path, key=key)

        if chunksize is None:
            n_chunks = 1
            chunksize = n_rows
        else:
            n_chunks = int(np.ceil(n_rows / chunksize))
            log.info('Splitting data into {} chunks'.format(n_chunks))

        for col in copy(columns):
            if group[col].ndim > 2:
                columns.remove(col)
                log.warning('Ignoring column {}, not 1d or 2d'.format(col))

        for chunk in range(n_chunks):

            start = chunk * chunksize
            end = min(n_rows, (chunk + 1) * chunksize)

            df = pd.DataFrame(index=np.arange(start, end))

            for col in columns:
                array = to_native_byteorder(group[col][start:end])

                # pandas cannot handle bytes, convert to str
                if array.dtype.kind == 'S':
                    array = array.astype(str)

                if array.ndim == 1:
                    df[col] = array

                else:
                    for i in range(array.shape[1]):
                        df[col + '_{}'.format(i)] = array[:, i]

            yield df, start, end


def read_data(file_path, key=None, columns=None, **kwargs):
    '''
    This is a utility wrapper for other reading functions.
    It will look for the file extension and try to use the correct
    reader and return a dataframe.

    Currently supported are hdf5 (pandas and h5py), json, jsonlines and csv

    Parameters
    ----------
    file_path: str
        Path to the input file

    All kwargs are passed to the individual reading function:

    pandas hdf5:   pd.read_hdf
    h5py hdf5:     fact.io.read_h5py
    json:          pd.DataFrame(json.load(file))
    jsonlines:     pd.read_json
    csv:           pd.read_csv
    '''
    name, extension = path.splitext(file_path)

    if extension in ['.hdf', '.hdf5', '.h5']:
        try:
            df = pd.read_hdf(file_path, key=key, columns=columns, **kwargs)
        except (TypeError, ValueError):
            df = read_h5py(file_path, key=key, columns=columns, **kwargs)
        return df

    if extension == '.json':
        with open(file_path, 'r') as j:
            d = json.load(j)
            df = pd.DataFrame(d)

    elif extension in ('.jsonl', '.jsonlines'):
        df = pd.read_json(file_path, lines=True, **kwargs)

    elif extension == '.csv':
        df = pd.read_csv(file_path, **kwargs)

    else:
        raise NotImplementedError('Unknown data file extension {}'.format(extension))

    if columns:
        df = df[columns]

    return df


def check_extension(file_path, allowed_extensions=allowed_extensions):
    p, extension = path.splitext(file_path)
    if extension not in allowed_extensions:
        raise IOError('Allowed formats: {}'.format(allowed_extensions))


def to_h5py(filename, df, key='data', mode='a', dtypes=None, index=True, **kwargs):
    '''
    Write pandas dataframe to h5py style hdf5 file

    Parameters
    ----------
    filename: str
        output file name
    df: pd.DataFrame
        The data to write out
    key: str
        the name for the hdf5 group to hold all datasets, default: data
    mode: str
        'w' to overwrite existing files, 'a' to append
    dtypes: dict
        if given, a mapping of column names to dtypes for conversion.
    index: bool
        If bool, also save the index of the dataframe

    All `**kwargs` are passed to h5py.create_dataset
    '''
    assert mode in ('w', 'a'), 'mode has to be either "a" or "w"'

    array = df.to_records(index=index)

    if dtypes is not None:
        array = change_recarray_dtype(array, dtypes)

    with h5py.File(filename, mode=mode) as f:
        if key not in f:
            initialize_h5py(f, array, key=key, **kwargs)

        append_to_h5py(f, array, key=key)


def change_recarray_dtype(array, dtypes):
    '''
    Change dtypes of recarray columns

    Parameters
    ----------
    array: np.recarray
      The array to modify
    dtypes: dict
        if given, a mapping of column names to dtypes for conversion.

    Returns
    -------
    array: np.recarray
        Copy of the original array with changed dtypes
    '''

    # get a modifieabale list of dtypes
    dt = array.dtype.descr
    # create a mapping from column name to index
    idx = {k: i for i, (k, d) in enumerate(dt)}

    for col, new_type in dtypes.items():
        dt[idx[col]] = (col, new_type)

    dt = np.dtype(dt)

    return array.astype(dt)


def initialize_h5py(f, array, key='events', **kwargs):
    '''
    Create a group with name `key` and empty datasets for each
    entry in dtypes.

    Parameters
    ----------
    f: h5py.File
        the hdf5 file, opened either in write or append mode
    array: numpy structured array
        The data
    key: str
        the name for the hdf5 group to hold all datasets, default: data

    All kwargs are passed to h5py create_dataset
    '''
    group = f.require_group(key)

    dtypes = array.dtype
    for name in dtypes.names:
        create_empty_h5py_dataset(array[name], group, name, **kwargs)

    return group


def create_empty_h5py_dataset(array, group, name, **kwargs):
    '''
    Create a new h5py dataset for the content of `array`.

    datetime64 objects are stored as fixed length strings
    arrays of lists are stored as 2d arrays (so they have to be fixed length)
    strings are stored as special dtype for strings

    Parameters
    ----------
    array: numpy.array or numpy.recarray
        the numpy array to get the dtype and shape from
    dataset: h5py.Group
        the hdf5 group the dataset should be created in
    name: str
        name for the new dataset
    **kwargs:
        all **kwargs are passed to create_dataset, useful for e.g. compression
    '''
    dtype = array.dtype
    maxshape = [None] + list(array.shape)[1:]
    shape = [0] + list(array.shape)[1:]

    if dtype.base == object:
        if isinstance(array[0], list):
            dt = np.array(array[0]).dtype
            shape = [0, len(array[0])]
            maxshape = [None, len(array[0])]
        else:
            dt = h5py.special_dtype(vlen=str)

    elif dtype.type == np.datetime64:
        # save dates as ISO string, create dummy date to get correct length
        dt = np.array(0, dtype=dtype).astype('S').dtype

    else:
        dt = dtype.base

    dataset = group.create_dataset(
        name,
        shape=tuple(shape),
        maxshape=tuple(maxshape),
        dtype=dt,
        **kwargs
    )
    return dataset


def append_to_h5py_dataset(array, dataset):
    '''
    Append a single numpy array to an h5py dataset.

    datetime64 objects are stored as fixed length strings
    arrays of lists are stored as 2d arrays (so they have to be fixed length)

    Parameters
    ----------
    array: numpy.array or numpy.recarray
        the numpy array to append
    dataset: h5py.Dataset
        the hdf5 dataset to append to
    '''
    n_existing_rows = dataset.shape[0]
    n_new_rows = array.shape[0]

    dataset.resize(n_existing_rows + n_new_rows, axis=0)

    # swap byteorder if not native
    if array.dtype.byteorder not in ('=', native_byteorder, '|'):
        data = array.newbyteorder().byteswap()
    else:
        data = array

    if data.dtype.type == np.datetime64:
        data = data.astype('S')

    if data.dtype.base == object:
        if isinstance(data[0], list):
            data = np.array([o for o in data])

    dataset[n_existing_rows:] = data


def append_to_h5py(f, array, key='events'):
    '''
    Append a numpy record or structured array to the given hdf5 file
    The file should have been previously initialized with initialize_hdf5

    Parameters
    ----------
    f: h5py.File
        the hdf5 file, opened either in write or append mode
    array: numpy.array or numpy.recarray
        the numpy array to append
    key: str
        the name for the hdf5 group with the corresponding data sets
    '''

    group = f.get(key)

    for column in array.dtype.names:
        dataset = group.get(column)
        if dataset is None:
            raise KeyError('No such dataset {}'.format(dataset))
        append_to_h5py_dataset(array[column], dataset)



import os.path
import re

__all__ = [
    'parse'
    'template_to_path',
    'tree_path',
]


def template_to_path(night, run, template, **kwargs):
    '''Make path from template and (night, run) using kwargs existing.

    night: int
        e.g. night = 20160102
        is used to create Y,M,D,N template values as:
        Y = "2016"
        M = "01"
        D = "02"
        N = "20160101"
    run: int or None
        e.g. run = 1
        is used to create template value R = "001"
    template: string
        e.g. "/foo/bar/{Y}/baz/{R}_{M}_{D}.gz.{N}"
    kwargs:
        if template contains other place holders than Y,M,D,N,R
        kwargs are used to format these.
    '''
    night = '{:08d}'.format(night)
    if run is not None:
        kwargs['R'] = '{:03d}'.format(run)

    kwargs['N'] = night
    kwargs['Y'] = night[0:4]
    kwargs['M'] = night[4:6]
    kwargs['D'] = night[6:8]
    return template.format(**kwargs)


def tree_path(night, run, prefix, suffix):
    '''Make a tree_path from a (night, run) for given prefix, suffix

    night: int
        eg. 20160101
    run: int or None
        eg. 11
    prefix: string
        eg. '/fact/raw' or '/fact/aux'
    suffix: string
        eg. '.fits.fz' or '.log' or '.AUX_FOO.fits'
    '''
    if run is not None:
        base_name = '{N}_{R}'
    else:
        base_name = '{N}'

    template = os.path.join(
            prefix,
            '{Y}',
            '{M}',
            '{D}',
            base_name + suffix)
    return template_to_path(night, run, template)


path_regex = re.compile(
    r'(?P<prefix>.*?)' +
    r'((/\d{4})(/\d{2})(/\d{2}))?/' +
    r'(?P<night>\d{8})' +
    r'(_?(?P<run>\d{3}))?' +
    r'(?P<suffix>.*)'
)


def parse(path):
    '''Return a dict with {prefix, suffix, night, run} parsed from path.

    path: string
        any (absolute) path should be fine.
    '''
    d = path_regex.match(path).groupdict()
    if d['run'] is not None:
        d['run'] = int(d['run'])
    d['night'] = int(d['night'])
    return d



from astropy.coordinates import (
    BaseCoordinateFrame,
    AltAz,
    frame_transform_graph,
    FunctionTransform,
)

try:
    from astropy.coordinates import CoordinateAttribute, TimeAttribute, EarthLocationAttribute
except ImportError:
    # for astropy <= 2.0.0
    from astropy.coordinates import (
        CoordinateAttribute,
        TimeFrameAttribute as TimeAttribute,
        EarthLocationAttribute,
    )

from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.coordinates.representation import CartesianRepresentation
import astropy.units as u

from .representation import PlanarRepresentation
from ..instrument.constants import FOCAL_LENGTH_MM, LOCATION
import numpy as np

focal_length = FOCAL_LENGTH_MM * u.mm


class CameraFrame(BaseCoordinateFrame):
    '''Astropy CoordinateFrame representing coordinates in the CameraPlane'''
    default_representation = PlanarRepresentation
    pointing_direction = CoordinateAttribute(frame=AltAz, default=None)
    obstime = TimeAttribute(default=None)
    location = EarthLocationAttribute(default=LOCATION)


@frame_transform_graph.transform(FunctionTransform, CameraFrame, AltAz)
def camera_to_altaz(camera, altaz):
    if camera.pointing_direction is None:
        raise ValueError('Pointing Direction must not be None')

    x = camera.x.copy()
    y = camera.y.copy()

    z = 1 / np.sqrt(1 + (x / focal_length)**2 + (y / focal_length)**2)
    x *= z / focal_length
    y *= z / focal_length

    cartesian = CartesianRepresentation(x, y, z, copy=False)

    rot_z_az = rotation_matrix(-camera.pointing_direction.az, 'z')
    rot_y_zd = rotation_matrix(-camera.pointing_direction.zen, 'y')

    cartesian = cartesian.transform(rot_y_zd)
    cartesian = cartesian.transform(rot_z_az)

    altitude = 90 * u.deg - np.arccos(cartesian.z)
    azimuth = np.arctan2(cartesian.y, cartesian.x)

    return AltAz(
        alt=altitude,
        az=azimuth,
        location=camera.location,
        obstime=camera.obstime,
    )


@frame_transform_graph.transform(FunctionTransform, AltAz, CameraFrame)
def altaz_to_camera(altaz, camera):
    if camera.pointing_direction is None:
        raise ValueError('Pointing Direction must not be None')

    cartesian = altaz.cartesian

    rot_z_az = rotation_matrix(camera.pointing_direction.az, 'z')
    rot_y_zd = rotation_matrix(camera.pointing_direction.zen, 'y')

    cartesian = cartesian.transform(rot_z_az)
    cartesian = cartesian.transform(rot_y_zd)

    return CameraFrame(
        x=cartesian.x * focal_length / cartesian.z,
        y=cartesian.y * focal_length / cartesian.z,
        pointing_direction=camera.pointing_direction,
        obstime=altaz.obstime,
        location=camera.location,
    )



'''
Constants describing the FACT telescope
'''
import numpy as np
from astropy.coordinates import EarthLocation
import astropy.units as u

#: The inner diameter of the hexagonal pixels in mm.
#: This is also the grid constant of the hex grid.
PIXEL_SPACING_MM = 9.5
FOCAL_LENGTH_MM = 4889
#:The segmented imaging reflector of FACT is well described using the thin lens
#:equation. However, the most prominent deviation from the thin lens, is the
#:imaging reflectors pincushin distortion. The pincushin distortion projects
#:incomning light further away from the optical axis as it is expected from the
#:thin lens equation. The additional outward pin-cushin distortion gets stronger
#:the further the distance to the optical axis is.
#:
#:``actual_projection_angle = (1 + PINCUSHION_DISTORTION_SLOPE) * thin_lens_prediction_angle``
#:
#:Example:
#:According to the tĥin lens model one expects incoming light of 1.5 deg
#:incident angle relative to the optical axis to be projected in:
#:
#:``tan(1.5deg)*focal_length = 128.02mm``
#:
#:distance to the optical axis on the image sensor screen.
#:But the reflector actually projects this incoming light further to the
#:outside to:
#:
#:``128.02mm * (1 + PINCUSHION_DISTORTION_SLOPE) = 130.67mm``
#:
#:As one can see, the correction is minor. It is only half a pixel at the outer
#:rim of the field of view.
#:
#:.. image:: figures/pincushin_distortion_slope.png
PINCUSHION_DISTORTION_SLOPE = 0.031/1.5
LATITUDE_DEC_DEG = 28.7616  #: FACT's latitude in decimal degrees
LONGITUDE_DEC_DEG = -17.8911   #: FACT's longitude in decimal degrees
ALTITUDE_ASL_M = 2200  #: FACT's altitude above sea level in meters
FOCAL_LENGTH_MM = 4889  #: FACT's reflector focal length in mm.
#: Field of view of a single pixel in decimal degrees
FOV_PER_PIXEL_DEG = np.rad2deg(2 * np.arctan(0.5 * PIXEL_SPACING_MM / FOCAL_LENGTH_MM))

#: astropy.coordinates.EarthLocation for the fact position
LOCATION = EarthLocation(
    lat=LATITUDE_DEC_DEG * u.deg, lon=LONGITUDE_DEC_DEG * u.deg,
    height=ALTITUDE_ASL_M * u.m
)



from setuptools import setup

with open('fact/VERSION', 'r') as f:
    __version__ = f.read().strip()


setup(
    name='pyfact',
    version=__version__,
    description='A module containing useful methods for working with fact',
    url='http://github.com/fact-project/pyfact',
    author='Maximilian Noethe, Dominik Neise',
    author_email='maximilian.noethe@tu-dortmund.de',
    license='MIT',
    packages=[
        'fact',
        'fact.plotting',
        'fact.slowdata',
        'fact.credentials',
        'fact.auxservices',
        'fact.factdb',
        'fact.analysis',
        'fact.instrument',
        'fact.coordinates',
    ],
    package_data={
        '': [
            'VERSION',
            'resources/*',
            'credentials/credentials.encrypted',
        ]
    },
    tests_require=['pytest>=3.0.0'],
    setup_requires=['pytest-runner'],
    install_requires=[
        'astropy',
        'h5py',
        'matplotlib>=1.4',
        'numpy',
        'pandas',
        'peewee',
        'pymongo>=2.7',
        'pymysql',
        'python-dateutil',
        'scipy',
        'setuptools',
        'simple-crypt',
        'sqlalchemy',
        'tables',  # pytables in anaconda
        'wrapt',
    ],
    zip_safe=False,
)



import numpy as np


def random_power(spectral_index, e_min, e_max, size):
    r'''
    Draw random numbers from a power law distribution

    .. math::
        f(E) =
        \frac{1 - \gamma}
        {E_{\max}^{1 - \gamma} - E_{\min}^{1 - \gamma}}
        E^{-\gamma}

    Parameters
    ----------
    spectral_index: float
        The differential spectral index of the power law
    e_min: float
        lower energy border
    e_max: float
        upper energy border, can be np.inf
    size: int or tuple[int]
        Number of events to draw or a shape like (100, 2)
    '''
    assert spectral_index > 1.0, 'spectral_index must be > 1.0'
    u = np.random.uniform(0, 1, size)

    exponent = 1 - spectral_index

    if e_max == np.inf:
        diff = -e_min**exponent
    else:
        diff = e_max**exponent - e_min**exponent

    return (diff * u + e_min**exponent) ** (1 / exponent)


def power_law(energy, flux_normalization, spectral_index):
    r'''
    Simple power law

    .. math::
        \phi = \phi_0 \cdot E ^{-\gamma}

    Parameters
    ----------
    energy: number or array-like
        energy points to evaluate
    flux_normalization: float
        Flux normalization
    spectral_index: float
        Spectral index
    '''
    return flux_normalization * energy**(-spectral_index)


def curved_power_law(energy, flux_normalization, a, b):
    r'''
    Curved power law

    .. math::
        \phi = \phi_0 \cdot E ^ {a - b \cdot \log(E)}

    Parameters
    ----------
    energy: number or array-like
        energy points to evaluate
    flux_normalization: float
        Flux normalization
    a: float
        Parameter `a`  of the curved power law
    b: float
        Parameter `b` of the curved power law
    '''
    return flux_normalization * energy ** (a + b * np.log10(energy))


def li_ma_significance(n_on, n_off, alpha=0.2):
    '''
    Calculate the Li&Ma significance for given
    observations data

    Parameters
    ----------
    n_on: integer or array like
        Number of events for the on observations
    n_off: integer of array like
        Number of events for the off observations
    alpha: float
        Scaling factor for the off observations, for wobble observations
        this is 1 / number of off regions
    '''

    scalar = np.isscalar(n_on)

    n_on = np.array(n_on, copy=False, ndmin=1)
    n_off = np.array(n_off, copy=False, ndmin=1)

    with np.errstate(divide='ignore', invalid='ignore'):
        p_on = n_on / (n_on + n_off)
        p_off = n_off / (n_on + n_off)

        t1 = n_on * np.log(((1 + alpha) / alpha) * p_on)
        t2 = n_off * np.log((1 + alpha) * p_off)

        ts = (t1 + t2)
        significance = np.sqrt(ts * 2)

    significance[np.isnan(significance)] = 0
    significance[n_on < alpha * n_off] = 0

    if scalar:
        return significance[0]

    return significance


def calc_weight_simple_to_curved(energy, spectral_index, a, b):
    '''
    Reweight simulated events from a simulated power law with
    spectral index `spectral_index` to a curved power law with parameters `a` and `b`

    Parameters
    ----------
    energy: float or array-like
        Energy of the events
    spectral_index: float
        Spectral index of the simulated power law
    a: float
        Parameter `a` of the target curved power law
    b: float
        Parameter `b` of the target curved power law
    '''
    return energy ** (spectral_index + a + b * np.log10(energy))


def calc_weight_change_index(energy, simulated_index, target_index):
    '''
    Reweight simulated events from one power law index to another

    Parameters
    ----------
    energy: float or array-like
        Energy of the events
    simulated_index: float
        Spectral index of the simulated power law
    target_index: float
        Spectral index of the target power law
    '''
    return energy ** (target_index - simulated_index)


def calc_gamma_obstime(n_events, spectral_index, flux_normalization, max_impact, e_min, e_max):
    '''
    Calculate the equivalent observation time for a spectral_index montecarlo set

    Parameters
    ----------
    n_events: int
        Number of simulated events
    spectral_index: float
        Spectral index of the simulated power law
    flux_normalization: float
        Flux normalization of the simulated power law
    max_impact: float
        Maximal simulated impact
    e_min: float
        Mimimal simulated energy
    e_max: float
        Maximal simulated energy
    '''
    if spectral_index >= -1:
        raise ValueError('spectral_index must be < -1')

    numerator = n_events * (1 - spectral_index)

    t1 = flux_normalization * max_impact**2 * np.pi
    t2 = e_max**(1 - spectral_index) - e_min**(1 - spectral_index)

    denominator = t1 * t2

    return numerator / denominator


def power_law_integral(flux_normalisation, spectral_index, e_min, e_max):
    '''
    Return the integral of a power_law with normalisation
    `flux_normalization` and index `spectral_index`
    between `e_min` and `e_max`
    '''
    int_index = 1 - spectral_index
    return flux_normalisation / int_index * (e_max**(int_index) - e_min**(int_index))


def calc_proton_obstime(
        n_events,
        spectral_index,
        max_impact,
        viewcone,
        e_min,
        e_max,
        flux_normalization=1.8e4,
        ):
    '''
    Calculate the equivalent observation time for a proton montecarlo set

    Parameters
    ----------
    n_events: int
        Number of simulated events
    spectral_index: float
        Spectral index of the simulated power law
    max_impact: float
        Maximal simulated impact in m
    viewcone: float
        Viewcone in degrees
    e_min: float
        Mimimal simulated energy in GeV
    e_max: float
        Maximal simulated energy in GeV
    flux_normalization: float
        Flux normalisation of the cosmic rays in nucleons / (m² s sr GeV)
        Default value is (29.2) of
        http://pdg.lbl.gov/2016/reviews/rpp2016-rev-cosmic-rays.pdf
    '''
    area = np.pi * max_impact**2
    solid_angle = 2 * np.pi * (1 - np.cos(np.deg2rad(viewcone)))

    expected_integral_flux = power_law_integral(flux_normalization, 2.7, e_min, e_max)

    return n_events / area / solid_angle / expected_integral_flux
